from __future__ import annotations

import os
import pickle
from tempfile import TemporaryDirectory
from typing import Any, Iterable, List, Tuple

import gdown
from PIL import Image
from torch import Tensor
from torch.utils.data import Dataset
from torchdata.datapipes.iter import IterableWrapper
from tqdm import tqdm

from clip_text_decoder.common import check_vision_backbone
from clip_text_decoder.datapipes import ParallelImageEncoder, coco_captions_datapipe

COCO_ANNOTATIONS_URL = (
    "http://images.cocodataset.org/annotations/annotations_trainval2014.zip"
)
CACHE_URLS = {
    "blip:base": {
        "train": "https://drive.google.com/uc?id=1vzmE9dyHeTyL0S7pYj8RurY82W53SGqE",
        # https://drive.google.com/file/d/1vzmE9dyHeTyL0S7pYj8RurY82W53SGqE/view?usp=sharing
        "val": "https://drive.google.com/uc?id=1aeUKoWOSjHd2K_SAKZUhmrzreXdiuydT",
        # https://drive.google.com/file/d/1aeUKoWOSjHd2K_SAKZUhmrzreXdiuydT/view?usp=sharing
    },
    "clip:ViT-B/32": {
        "train": "https://drive.google.com/uc?id=1e-K7UIgVsvsHZEkZguzhTqEoMAUfu538",
        # https://drive.google.com/file/d/1e-K7UIgVsvsHZEkZguzhTqEoMAUfu538/view?usp=sharing
        "val": "https://drive.google.com/uc?id=11l6b9rol53FAZe4EhvlgiwrsD4qfUUdj",
        # https://drive.google.com/file/d/11l6b9rol53FAZe4EhvlgiwrsD4qfUUdj/view?usp=sharing
    },
}


class CachedDataset(Dataset):
    def __init__(self, data: List[Tuple[Tensor, Any]]):
        super().__init__()
        self.data = data

    @classmethod
    def load(cls, path: str) -> CachedDataset:
        with open(path, "rb") as f:
            data = pickle.load(f)
        return CachedDataset(data=data)

    def save(self, path: str):
        with open(path, "wb") as f:
            pickle.dump(self.data, f)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Tuple[Tensor, str]:
        return self.data[idx]


def build_cached_dataset(
    data_iterable: Iterable[Image.Image, Any],
    vision_backbone: str,
) -> CachedDataset:
    print("Compiling encodings with text captions...")
    pipe = IterableWrapper(iter(data_iterable))
    pipe = pipe.batch(32)
    pipe = ParallelImageEncoder(pipe, vision_backbone=vision_backbone)
    pipe = pipe.unbatch()

    data = [(encoding, captions) for encoding, captions in tqdm(pipe)]
    return CachedDataset(data=data)


class CocoCaptionsDataset(CachedDataset):
    @staticmethod
    def download(url: str) -> CocoCaptionsDataset:
        with TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, "cache.pkl")
            gdown.download(url, path, quiet=False)
            with open(path, "rb") as f:
                data = pickle.load(f)

        return CocoCaptionsDataset(data=data)

    @classmethod
    def build(
        cls,
        vision_backbone: str = "blip:base",
        root: str = "./coco-captions",
        split: str = "train",
        force_rebuild: bool = False,
    ) -> CocoCaptionsDataset:
        check_vision_backbone(vision_backbone)
        if force_rebuild or vision_backbone not in CACHE_URLS:
            pipe = coco_captions_datapipe(cache_dir=root, split=split)
            cached_dataset = build_cached_dataset(pipe, vision_backbone=vision_backbone)
            return CocoCaptionsDataset(data=cached_dataset.data)
        else:
            url_by_split = CACHE_URLS[vision_backbone]
            url = url_by_split[split]
            return cls.download(url)



from __future__ import annotations

import asyncio
import io
import json
import os
from collections import defaultdict
from typing import Dict, Generator, List, Sequence, Tuple
from zipfile import ZipFile

import numpy as np
import torch
import wget
from PIL import Image
from torchdata.datapipes.iter import IterableWrapper, IterDataPipe
from tqdm import tqdm

from clip_text_decoder.common import encode_image_tensor, load_vision_backbone
from clip_text_decoder.utils.fileio import async_batch_get_request

COCO_ANNOTATIONS_URL = (
    "http://images.cocodataset.org/annotations/annotations_trainval2014.zip"
)
TRAIN_JSON = "annotations/captions_train2014.json"
VAL_JSON = "annotations/captions_val2014.json"


def _download_coco_captions_json(cache_dir: str):
    zip_path = os.path.join(cache_dir, "annotations.zip")
    if not os.path.exists(zip_path):
        print("Downloading COCO Captions annotations.")
        wget.download(COCO_ANNOTATIONS_URL, zip_path)

    with ZipFile(zip_path, mode="r") as zip:
        if not os.path.exists(os.path.join(cache_dir, TRAIN_JSON)):
            zip.extract(TRAIN_JSON, path=cache_dir)
        if not os.path.exists(os.path.join(cache_dir, VAL_JSON)):
            zip.extract(VAL_JSON, path=cache_dir)


def _load_coco_captions_json(cache_dir: str, split: str) -> Dict[str, List]:
    _download_coco_captions_json(cache_dir)

    if split == "train":
        json_file = TRAIN_JSON
    elif split == "val":
        json_file = VAL_JSON
    else:
        raise ValueError(f"Invalid split '{split}'. Available: ['train', 'val'].")

    with open(os.path.join(cache_dir, json_file)) as f:
        return json.load(f)


def _get_captions_by_image_id(annotations: Sequence[Dict]) -> Dict[int, List[str]]:
    out: Dict[int, List[str]] = defaultdict(list)
    for ann in annotations:
        image_id = ann["image_id"]
        out[image_id].append(ann["caption"])

    return out


class ParallelSampleDownloader(IterDataPipe):
    def __init__(self, dp: IterDataPipe[str]) -> None:
        super().__init__()
        self.dp = dp

    def __iter__(self) -> Generator[Image.Image, None, None]:
        for batch in self.dp:
            images = asyncio.run(async_batch_get_request([x[0] for x in batch]))
            captions = [x[1] for x in batch]

            batch_results = []
            for _image, _captions in zip(images, captions):
                if _image is None:
                    continue
                try:
                    image = Image.open(io.BytesIO(_image))
                    batch_results.append((image, _captions))
                except Exception:
                    continue

            yield batch_results


class ParallelImageEncoder(IterDataPipe):
    def __init__(
        self, dp: IterDataPipe[Tuple[Image.Image, List[str]]], vision_backbone: str
    ):
        super().__init__()
        self.dp = dp
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model, self.preprocessor = load_vision_backbone(
            vision_backbone, device=self.device
        )

    @torch.inference_mode()
    @torch.no_grad()
    def __iter__(self) -> Generator[Tuple[np.ndarray, List[str]], None, None]:
        for batch in self.dp:
            captions = [caps for _, caps in batch]
            images = torch.stack(
                [self.preprocessor(image.convert("RGB")) for image, _ in batch]
            ).to(self.device)
            image_features = encode_image_tensor(images, self.model)

            # For legacy reasons, unsqueeze along dimension 1.
            image_features_np = image_features.unsqueeze(1).cpu().numpy()
            yield [(feats, caps) for feats, caps in zip(image_features_np, captions)]


def coco_captions_datapipe(
    cache_dir: str = "./coco-captions",
    split: str = "train",
    buffer_size: int = 128,
) -> IterDataPipe:
    captions_json = _load_coco_captions_json(cache_dir=cache_dir, split=split)
    images, annotations = captions_json["images"], captions_json["annotations"]
    captions_by_image_id = _get_captions_by_image_id(annotations)
    images_with_captions = [
        (image["coco_url"], captions_by_image_id[image["id"]]) for image in images
    ]

    pipe: IterDataPipe = IterableWrapper(images_with_captions)
    pipe = pipe.batch(buffer_size)
    pipe = ParallelSampleDownloader(pipe)
    pipe = pipe.unbatch()

    return pipe


if __name__ == "__main__":
    pipe = coco_captions_datapipe()

    for _ in tqdm(pipe.batch(32)):
        pass



from __future__ import annotations

from enum import Enum
from typing import Callable, Optional, Tuple, Union

import clip
import torch
from clip.model import CLIP
from lavis.models import BlipFeatureExtractor, load_model_and_preprocess
from PIL import Image
from torch import Tensor, nn
from transformers import GPT2Config, GPT2LMHeadModel, GPT2Tokenizer

PreprocessorType = Callable[[Image.Image], Tensor]


class LanguageModels(Enum):
    distilgpt2: str = "distilgpt2"
    gpt2: str = "gpt2"
    gpt2_medium: str = "gpt2-medium"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def check_language_model(name: str) -> None:
    allowed = LanguageModels.list()
    if name not in allowed:
        raise ValueError(f"Unsupported language model '{name}'. Allowed: {allowed}.")


def load_language_model(
    name: str, device: Optional[Union[str, torch.device]] = None
) -> nn.Module:
    check_language_model(name)
    config = GPT2Config.from_pretrained(name, add_cross_attention=True)
    model = GPT2LMHeadModel.from_pretrained(name, config=config)
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    model = model.to(device)
    return model


def load_tokenizer(name: str) -> GPT2Tokenizer:
    check_language_model(name)
    tokenizer = GPT2Tokenizer.from_pretrained(name)
    tokenizer.pad_token = tokenizer.eos_token
    return tokenizer


class VisionBackbones(Enum):
    blip_base: str = "blip:base"
    clip_rn50: str = "clip:RN50"
    clip_rn101: str = "clip:RN101"
    clip_vit_b32: str = "clip:VIT-B/32"
    clip_vit_b16: str = "clip:VIT-B/16"
    clip_vit_l14: str = "clip:VIT-L/14"
    clip_vit_l14_336px: str = "clip:VIT-L/14@336px"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def check_vision_backbone(backbone: str) -> None:
    allowed = VisionBackbones.list()
    if backbone not in allowed:
        raise ValueError(f"Unsupported backbone '{backbone}'. Allowed: {allowed}.")


def load_vision_backbone(
    backbone: str, device: Optional[Union[str, torch.device]] = None
) -> Tuple[nn.Module, PreprocessorType]:
    check_vision_backbone(backbone)
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    if backbone == VisionBackbones.blip_base.value:
        model, preprocessors, _ = load_model_and_preprocess(
            "blip_feature_extractor", model_type="base", is_eval=True, device=device
        )
        return model, preprocessors["eval"]
    else:
        # Currently, all other supported backbones are CLIP
        _, name = backbone.split(":")
        return clip.load(name, device=device, jit=False)


def encode_image_tensor(image: Tensor, backbone: nn.Module) -> Tensor:
    if isinstance(backbone, BlipFeatureExtractor):
        features = backbone.extract_features({"image": image}, mode="image")
        return features.image_embeds[:, 0]
    else:
        # Currently, all other supported backbones are CLIP
        assert isinstance(backbone, CLIP)
        return backbone.encode_image(image)



import asyncio
from typing import Optional, Sequence

import aiohttp


async def _async_get_request(
    session: aiohttp.ClientSession, url: str
) -> Optional[bytes]:
    try:
        resp = await session.get(url)
        return await resp.read()
    except Exception:
        return None


async def async_batch_get_request(urls: Sequence[str], timeout: float = 1.0):
    client_timeout = aiohttp.ClientTimeout(total=timeout)
    async with aiohttp.ClientSession(timeout=client_timeout) as session:
        return await asyncio.gather(*[_async_get_request(session, url) for url in urls])



import os
from distutils.core import setup
from subprocess import getoutput

import setuptools


def get_version_tag() -> str:
    try:
        version = os.environ["CLIP_TEXT_DECODER_VERSION"]
    except KeyError:
        version = getoutput("git describe --tags --abbrev=0")

    return version


setup(
    name="clip-text-decoder",
    version=get_version_tag(),
    author="Frank Odom",
    author_email="frank.odom.iii@gmail.com",
    url="https://github.com/fkodom/clip-text-decoder",
    packages=setuptools.find_packages(exclude=["tests"]),
    description="Generate text captions for images from their CLIP embeddings.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "evaluate",
        "gdown",
        "numpy",
        "pytorch-lightning",
        "spacy",
        "torch>=1.11",
        "torchdata>=0.3.0",
        "torchvision",
        "transformers",
        "wget",
    ],
    extras_require={
        "test": [
            "black",
            "flake8",
            "isort",
            "pre-commit",
            "pytest",
            "pytest-cov",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



from __future__ import annotations

import os
import tempfile
from typing import Callable, List, Optional, Tuple, Union

import gdown
import torch
import torch.nn.functional as F
from PIL import Image
from pytorch_lightning import LightningModule
from torch import Tensor, nn, optim
from transformers import GPT2Tokenizer

from clip_text_decoder.common import (
    check_language_model,
    check_vision_backbone,
    encode_image_tensor,
    load_language_model,
    load_vision_backbone,
)

PRETRAINED_INFERENCE_MODEL_PATH = (
    "https://drive.google.com/uc?id=1bEAyV2279C4V4iYMaJahREiM58vjy6G1"
    # https://drive.google.com/file/d/1bEAyV2279C4V4iYMaJahREiM58vjy6G1/view?usp=sharing
)


class Decoder(LightningModule):
    def __init__(
        self,
        vision_backbone: str = "blip:base",
        language_model: str = "distilgpt2",
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        self.save_hyperparameters()
        check_vision_backbone(vision_backbone)
        self.vision_backbone = vision_backbone
        check_language_model(language_model)
        self.language_model = load_language_model(language_model, device=device)

        self.to(device)

    def forward(
        self,
        input_ids: Tensor,
        encoder_hidden_states: Tensor,
        attention_mask: Optional[Tensor] = None,
        labels: Optional[Tensor] = None,
    ):
        batch_size, _, num_features = encoder_hidden_states.shape
        # TODO: Check if we can get '768' (num_features) from the GPT2 model.
        hidden = torch.zeros(
            size=(batch_size, 1, 768),
            dtype=encoder_hidden_states.dtype,
            device=encoder_hidden_states.device,
        )
        hidden[:, :, :num_features] = encoder_hidden_states

        return self.language_model(
            input_ids=input_ids,
            encoder_hidden_states=hidden,
            attention_mask=attention_mask,
            labels=labels,
        )

    def configure_optimizers(self):
        return optim.AdamW(self.parameters(), lr=1e-4)

    def training_step(self, batch: Tuple[Tensor, Tensor, Tensor], *_) -> Tensor:
        encoder_hidden_states, input_ids, attention_mask = batch
        result = self.forward(
            input_ids=input_ids,
            encoder_hidden_states=encoder_hidden_states,
            attention_mask=attention_mask,
            labels=input_ids,
        )

        self.log(
            "training_loss",
            result.loss,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        return result.loss

    @torch.no_grad()
    def validation_step(self, batch: Tuple[Tensor, Tensor, Tensor], *_) -> Tensor:
        encoder_hidden_states, input_ids, attention_mask = batch
        result = self.forward(
            input_ids=input_ids,
            encoder_hidden_states=encoder_hidden_states,
            attention_mask=attention_mask,
            labels=input_ids,
        )

        self.log(
            "validation_loss",
            result.loss,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        return result.loss


class DecoderInferenceModel:
    _model_path = "model.pt"
    _tokenizer_path = "tokenizer.pkl"

    def __init__(
        self,
        model: Decoder,
        tokenizer: GPT2Tokenizer,
    ):
        self.model = model.eval()
        self.tokenizer = tokenizer

    @property
    def device(self) -> torch.device:
        return next(self.model.parameters()).device

    def to(self, device: torch.device) -> DecoderInferenceModel:
        self.model.to(device)
        return self

    def save(self, path: str):
        # Save a copy of the current model weights, and cast to FP16 for storage
        model_state_dict = self.model.state_dict()
        # Avoid saving any cached properties of this class or its subclasses :)
        obj = self.__class__(model=self.model.half(), tokenizer=self.tokenizer)
        torch.save(obj, path)
        # Restore the original model weights
        self.model.load_state_dict(model_state_dict)

    @classmethod
    def load(cls, path: str) -> DecoderInferenceModel:
        temp = torch.load(path)
        # Just in case we change any of the class methods here, unpack the model
        # and tokenizer, and pass them into a new instance of this class.
        return cls(model=temp.model.float(), tokenizer=temp.tokenizer)

    @classmethod
    def download_pretrained(cls, dest: str = None) -> DecoderInferenceModel:
        with tempfile.TemporaryDirectory() as tempdir:
            if dest is None:
                dest = os.path.join(tempdir, "model.zip")
            gdown.download(PRETRAINED_INFERENCE_MODEL_PATH, dest, quiet=False)
            return cls.load(dest)

    @torch.cuda.amp.autocast()
    @torch.no_grad()
    def __call__(self, x: Tensor, max_len: int = 64, beam_size: int = 1) -> str:
        """Inference using beam search. For beam search, we predict one token per step.
        After each step, we keep only the 'beam_size' output sequences with the highest
        end-to-end confidence score. Repeat this process until at most 'max_len' tokens
        have been generated.
        """
        encoder_hidden_states = x.reshape(1, 1, -1).to(self.device)
        # Since we haven't performed any beam search steps yet, we just have one
        # set of input IDs (with a single "start" token). We use 'None' for the log
        # probability of this sequence, since it's not being predicted by the model.
        input_ids = [torch.tensor([self.tokenizer.bos_token_id], device=self.device)]
        beam_logprobs: Optional[List[float]] = None

        def _get_beam_outputs(_input_ids: Tensor) -> Tuple[List[Tensor], Tensor]:
            """Performs inference on the 'input_ids' Tensor, and collects the top
            'beam_size' results by score. Returns a list of output Tensors, and
            their respective log-probabilities.
            """
            outputs = self.model(_input_ids.unsqueeze(0), encoder_hidden_states)
            logits: Tensor = outputs.logits[0, -1]
            logprobs = F.log_softmax(logits, dim=-1)

            topk_logprobs = logprobs.topk(k=beam_size)
            indices = topk_logprobs.indices
            logprobs = topk_logprobs.values
            output_ids = [
                torch.cat([_input_ids, idx.reshape(-1)], dim=0) for idx in indices
            ]

            return output_ids, logprobs

        for _ in range(max_len - 1):
            output_ids: List[Tensor] = []
            logprobs: List[float] = []
            beams_done: List[bool] = []

            # Collect the top 'beam_size' results from each beam individually.
            for beam_idx, ids in enumerate(input_ids):
                # If 'beam_logprobs' is already defined, then we've predicted at least
                # one token already. And if the last token is equal to the "stop" token,
                # we don't need to perform inference with this beam anymore.
                if beam_logprobs and ids[-1].item() == self.tokenizer.eos_token_id:
                    output_ids.append(ids)
                    logprobs.append(beam_logprobs[beam_idx])
                    beams_done.append(True)
                    continue

                _output_ids, _logprobs = _get_beam_outputs(ids)
                if beam_logprobs is not None:
                    # Sum the log-probabilities of the existing beam and our predicted
                    # token to get the total log-probability.
                    _logprobs += beam_logprobs[beam_idx]

                # Append the results from this beam to the aggregate lists.
                output_ids += _output_ids
                logprobs += _logprobs.tolist()
                beams_done.append(False)

            if all(beams_done):
                # All search beams are done generating text.
                break

            # Keep only the top 'beam_size' beams by total log-probability.
            indices = torch.tensor(logprobs).topk(k=beam_size).indices
            input_ids = [output_ids[idx] for idx in indices]
            beam_logprobs = [logprobs[idx] for idx in indices]

        # Find the predicted beam with highest overall log-probability.
        best_beam_idx: int = torch.tensor(beam_logprobs).argmax().item()  # type: ignore
        # Decode the predicted token IDs into a text string.
        return self.tokenizer.decode(input_ids[best_beam_idx], skip_special_tokens=True)


class ImageCaptionInferenceModel(DecoderInferenceModel):
    def __init__(self, model: Decoder, tokenizer: GPT2Tokenizer):
        super().__init__(model, tokenizer)
        self._vision_backbone: Optional[nn.Module] = None
        self._preprocessor: Optional[Callable] = None

    def _load_vision_backbone(self):
        backbone, preprocessor = load_vision_backbone(self.model.vision_backbone)
        self._vision_backbone = backbone
        self._preprocessor = preprocessor

    @property
    def vision_backbone(self) -> nn.Module:
        if self._vision_backbone is None:
            self._load_vision_backbone()
        assert self._vision_backbone is not None
        return self._vision_backbone

    @property
    def preprocessor(self) -> Callable:
        if self._preprocessor is None:
            self._load_vision_backbone()
        assert self._preprocessor is not None
        return self._preprocessor

    @torch.cuda.amp.autocast()
    @torch.no_grad()
    def __call__(
        self,
        image: Union[str, Image.Image],
        max_len: int = 64,
        beam_size: int = 1,
    ) -> str:
        if isinstance(image, str):
            image = Image.open(image)

        preprocessed: Tensor = self.preprocessor(image).to(self.device)
        encoded = encode_image_tensor(preprocessed.unsqueeze(0), self.vision_backbone)
        return super().__call__(encoded, max_len=max_len, beam_size=beam_size)
